unit SPI_MCP3208; 
{$mode objfpc}{$H+}
interface
uses
  Classes,Forms,LCLType,Dialogs,SysUtils,BaseUnix, Unix;

const
// SPI MCP3208 設定
    SPI_DEVICE      = '/dev/spidev0.0';  // SPI デバイス名

//    SPI_SPEED       = 1000000;           // Clock 1MHz
    SPI_SPEED       = 2000000;           // Clock 2 MHz

    SPI_BITS        = 8;                 // SPI 8bit設定
    SPI_DELAY       = 0;                 // CE0 delay タイミング

    MCP3208_START_BIT    = (1 shl 2);    // ADC MCP3208設定
    MCP3208_SINGLE_ENDED = (1 shl 1);   
    MCP3208_DIFFERENTIAL = (0 shl 1);
    MCP3208_CH_SHIFT     = 6;

//  /usr/include/linux/spi/spidev.h で定義されたもの
//  直接インクルードできないため書き直し
    SPI_IOC_MAGIC   = 'k';
    SPI_CPHA        = $01;
    SPI_CPOL        = $02;
    SPI_MODE_0      = $00;
    SPI_MODE_1      = SPI_CPHA;
    SPI_MODE_2      = SPI_CPOL;
    SPI_MODE_3      = SPI_CPOL or SPI_CPHA;
    SPI_CS_HIGH     = $04;
    SPI_LSB_FIRST   = $08;
    SPI_3WIRE       = $10;
    SPI_LOOP        = $20;
    SPI_NO_CS       = $40;
    SPI_READY       = $80;

    _IOC_NONE       =$00;
    _IOC_WRITE      =$01;
    _IOC_READ       =$02;
    _IOC_NRBITS     =  8;
    _IOC_TYPEBITS   =  8;
    _IOC_SIZEBITS   = 14;
    _IOC_DIRBITS    =  2;
    _IOC_NRSHIFT    =  0;
    _IOC_TYPESHIFT  = (_IOC_NRSHIFT+  _IOC_NRBITS);
    _IOC_SIZESHIFT  = (_IOC_TYPESHIFT+_IOC_TYPEBITS);
    _IOC_DIRSHIFT   = (_IOC_SIZESHIFT+_IOC_SIZEBITS);

type
  float = Single;

  spi_ioc_transfer_t =record  // sizeof(spi_ioc_transfer_t)= 32bytes
     tx_buf_ptr   :QWORD; // 8byte=64bit
     rx_buf_ptr   :QWORD; // 8byte=64bit
     len          :DWORD; // 4byte=32bit
     speedhz      :DWORD; // 4byte=32bit
     delay_usecs  :WORD;  // 2byte=16bit
     bits_per_word:BYTE;  // 1byte= 8bit
     cs_change    :BYTE;  // 1byte= 8bit
     pad          :DWORD; // 4byte=32bit
  end;

 TSPI = class(TObject)
      constructor Create;// procedure spi_init;
      destructor  Destroy;
      procedure   Read(var dt:Array of WORD);
      procedure   Read(var dt:Array of Single);
      procedure   Read(var dt:Single;ch:BYTE);
      procedure   Read(var dt:WORD;ch:BYTE);
      function    ShowStruct(spi_strct:spi_ioc_transfer_t):AnsiString;
  var
      ref_voltage:Single;     // ADC 1ビットに対する基準電圧を設定する。
      spi_fd:integer;
      mc3208_mode:Byte;
      tr:Array[0..7]  of spi_ioc_transfer_t;
      tx:Array[0..23] of BYTE;
      rx:Array[0..23] of BYTE;

      SPI_IOC_RD_MODE:         DWORD;// 32bit
      SPI_IOC_WR_MODE:         DWORD;
      SPI_IOC_RD_LSB_FIRST:    DWORD;
      SPI_IOC_WR_LSB_FIRST:    DWORD;
      SPI_IOC_RD_BITS_PER_WORD:DWORD;
      SPI_IOC_WR_BITS_PER_WORD:DWORD;
      SPI_IOC_RD_MAX_SPEED_HZ: DWORD;
      SPI_IOC_WR_MAX_SPEED_HZ: DWORD;
 end;

implementation
//********************************************
function SPI_MSGSIZE(n:byte):Word;
  var siz:Word;
begin
  if n*sizeof(spi_ioc_transfer_t) < (1 shl _IOC_SIZEBITS)
    then
       siz:=n*sizeof(spi_ioc_transfer_t)
    else
       siz:=0;
  SPI_MSGSIZE:=siz;
end;
//********************************************
function  _IOC(dir:byte; typ:char; nr,size:WORD):DWORD;
begin
  _ioc:=(dir      shl _IOC_DIRSHIFT)  or
        (ord(typ) shl _IOC_TYPESHIFT) or
        (nr       shl _IOC_NRSHIFT)   or
        (size     shl _IOC_SIZESHIFT);
end;
//********************************************
function _IO  (typ:char; nr:WORD):DWORD;     begin _IO  :=_IOC(_IOC_NONE,                typ,nr,   0);   end;
function _IOR (typ:char; nr,size:WORD):DWORD;begin _IOR :=_IOC(_IOC_Read,                typ,nr,size);   end;
function _IOW (typ:char; nr,size:WORD):DWORD;begin _IOW :=_IOC(_IOC_Write,               typ,nr,size);   end;
function _IOWR(typ:char; nr,size:WORD):DWORD;begin _IOWR:=_IOC((_IOC_Write or _IOC_Read),typ,nr,size);   end;
function SPI_IOC_MESSAGE(n:byte):DWORD;      begin SPI_IOC_MESSAGE:=_IOW(SPI_IOC_MAGIC,0,SPI_MSGSIZE(n));end;

//********************************************
destructor TSPI.Destroy;
begin
   if(spi_fd > 0) then 
      fpclose(spi_fd);
end;

//********************************************
//procedure TSPI.Init;
constructor TSPI.Create;
var
      ch:BYTE;
      ch_base:BYTE;
begin
        ref_voltage:=3.3/4095; // 1bit あたりの電圧
        
        SPI_IOC_RD_MODE :=          _IOR(SPI_IOC_MAGIC, 1, 1);
        SPI_IOC_WR_MODE :=          _IOW(SPI_IOC_MAGIC, 1, 1);
        SPI_IOC_RD_LSB_FIRST :=     _IOR(SPI_IOC_MAGIC, 2, 1);
        SPI_IOC_WR_LSB_FIRST :=     _IOW(SPI_IOC_MAGIC, 2, 1);
        SPI_IOC_RD_BITS_PER_WORD := _IOR(SPI_IOC_MAGIC, 3, 1);
        SPI_IOC_WR_BITS_PER_WORD := _IOW(SPI_IOC_MAGIC, 3, 1);
        SPI_IOC_RD_MAX_SPEED_HZ  := _IOR(SPI_IOC_MAGIC, 4, 4);
        SPI_IOC_WR_MAX_SPEED_HZ  := _IOW(SPI_IOC_MAGIC, 4, 4);
        for ch:=0 to 7 do
           begin
            ch_base:=ch*3;
             // tx is transmit data of  channel select
             tx[0+ch_base]:=MCP3208_START_BIT or MCP3208_SINGLE_ENDED or (ch shr 2) ;
             tx[1+ch_base]:=ch shl MCP3208_CH_SHIFT;
             tx[2+ch_base]:=0;

             tr[ch].tx_buf_ptr    :=QWORD(addr(tx[ch_base]));
             tr[ch].rx_buf_ptr    :=QWORD(addr(rx[ch_base]));

             tr[ch].len           :=3;
             tr[ch].speedhz       :=SPI_SPEED;
             tr[ch].delay_usecs   :=SPI_DELAY;
             tr[ch].bits_per_WORD :=SPI_BITS;
             tr[ch].cs_change     :=1; // must be 1 for 8ch read
           end;

        mc3208_mode:=SPI_MODE_3;
        spi_fd:=fpopen(SPI_DEVICE,O_RDWR);
        if(spi_fd = -1) then
           begin
               ShowMessage('spi device open error, check device driver');
               Exit;
          end;

        // Set SPI write mode
        if( fpioctl(spi_fd,SPI_IOC_WR_MODE,@mc3208_mode) < 0 )then
           begin
             ShowMessage('mcp3208 write set error');
             Exit;
           end;
        // Set SPI read mode
        if( fpioctl(spi_fd,SPI_IOC_RD_MODE,@mc3208_mode) < 0 )then
           begin
             ShowMessage('mcp3208 read set error ');
             Exit;
           end;
end;

//********************************************
//  ADCポートまとめ読み。引数は 1～chまで
//
//  dtが配列の場合 1～ch までを読込み配列に代入する。
//  2ch,4ch,8chとか、まとめて読みたい場合を想定している。
//********************************************
procedure  TSPI.Read(var dt:Array of WORD);
var
    i:BYTE;
    ch:BYTE;
begin
    ch:=SizeOf(dt);
    ch:=ch div 2; // WORDは2バイト
    if ch > 8 then ch:=8;
    
    fpioctl(spi_fd,SPI_IOC_MESSAGE(ch),addr(tr[0]));
    
    for i:=0 to ch-1 do
        begin
           dt[i]:=((rx[1+3*i] and $0F) shl 8) + rx[2+3*i];
        end;
//デバッグ出力用:構造体表示
//      ShowMessage(ShowStruct(tr[0]));

end;

//********************************************
//  ADCポートまとめ読み。
//
//  dtが配列の場合 1～ch までを読込み配列に代入する。
//  2ch,4ch,8chとか、まとめて読みたい場合を想定している。
//  複数読みの場合、0chから読まれる。
//********************************************
procedure  TSPI.Read(var dt:Array of Single);
var
    i:BYTE;
    ch:BYTE;
    val:WORD;
begin
    ch:=SizeOf(dt);
    ch:=ch div 4;   //  配列数を計算 Single( float相当)は４バイト
    if ch > 8 then ch:=8;
    
    fpioctl(spi_fd,SPI_IOC_MESSAGE(ch),addr(tr[0]));
    for i:=0 to ch-1 do
        begin
           dt[i]:=Single(((rx[1+3*i] and $0F) shl 8) + rx[2+3*i])*REF_VOLTAGE;
        end;

//デバッグ出力用:構造体表示
//      ShowMessage(ShowStruct(tr[0]));
       
end;

//********************************************
// 引数ch で指定したADCポートを1ch Singleで読み込む
//********************************************
procedure  TSPI.Read(var dt:Single;ch:BYTE);
var
    val:WORD;
begin

    if ch > 7 then ch:=7;
    fpioctl(spi_fd,SPI_IOC_MESSAGE(1),addr(tr[ch]));
    dt:=Single(((rx[1+3*ch] and $0F) shl 8) + rx[2+3*ch])*REF_VOLTAGE;

//デバッグ出力用:構造体表示
//      ShowMessage(ShowStruct(tr[0]));
end;

//********************************************
// 引数ch で指定したADCポートを1ch WORDで読み込む
//********************************************
procedure  TSPI.Read(var dt:WORD;ch:BYTE);
begin
    if ch > 7 then ch:=7;
    fpioctl(spi_fd,SPI_IOC_MESSAGE(1),addr(tr[ch]));
    dt:=((rx[1+3*ch] and $0F) shl 8) + rx[2+3*ch];
    
//デバッグ出力用:構造体表示
//        ShowMessage(ShowStruct(tr[ch]));
end;


//********************************************
function TSPI.ShowStruct(spi_strct:spi_ioc_transfer_t):AnsiString;
   var
       sstr:AnsiString;
begin
  with spi_strct do
  begin
    sstr:=    'SPI Struct     : 0x'+IntToHex(longword(addr(spi_strct)),8)
            +':'+IntToStr(sizeof(spi_strct))+'bytes'+#13#10
            +' .tx_buf_ptr   : 0x'+IntToHex(tx_buf_ptr,16)+':'+IntToStr(sizeof(tx_buf_ptr))+'bytes'+#13#10
            +' .rx_buf_ptr   : 0x'+IntToHex(rx_buf_ptr,16)+':'+IntToStr(sizeof(rx_buf_ptr))+'bytes'+#13#10
            +' .len          : '+IntToStr(len)+':'+IntToStr(sizeof(len))+'bytes'+#13#10
            +' .speedhz      : '+IntToStr(speedhz)+':'+IntToStr(sizeof(speedhz))+'bytes'+#13#10
            +' .delay_usecs  : '+IntToStr(delay_usecs)+':'+IntToStr(sizeof(delay_usecs))+'bytes'+#13#10
            +' .bits_per_word: '+IntToStr(bits_per_word)+':'+IntToStr(sizeof(bits_per_word))+'bytes'+#13#10
            +' .cs_change    : '+IntToStr(cs_change)+':'+IntToStr(sizeof(cs_change))+'bytes'+#13#10
            +' .pad          : 0x'+IntToHex(pad,8)+':'+IntToStr(sizeof(pad))+'bytes'+#13#10#10;
  end;
Result:=sstr;
end;
//********************************************
end.
